{$i replspydefs.pas}

unit replspyconndialog;

interface

uses SysUtils, Classes, Controls, StdCtrls, Forms, Graphics;

type
  TConnStringDialog = class(TForm)
    OKBtn: TButton;
    CancelBtn: TButton;
    ParamsCombo: TComboBox;
    OpenEdit: TEdit;
    OpenLbl: TLabel;
    FeatureCheck: TCheckBox;
  private
    { Private declarations }
  public
    procedure LoadItems(const afilename: string);
    procedure SaveItems(const afilename, aitem: string);
    procedure SetOpenParams(const adefault: string; aenabled: boolean);
  end;

function GetStdDialogParams(const acfgname, acaption, acheckcaption: string;
                            aenableopen, aenablecancel, aenableckeck: boolean;
                            var aopenparams: string; var cancelled, ischecked: boolean): string;

implementation

{$R *.dfm}

procedure TConnStringDialog.LoadItems(const afilename: string);
begin
  ParamsCombo.Items.Clear;
  if fileexists(afilename) then ParamsCombo.Items.LoadFromFile(afilename);
  if ParamsCombo.Items.Count > 0 then ParamsCombo.ItemIndex:= 0
                                 else ParamsCombo.Text:= '';
end;

procedure TConnStringDialog.SaveItems(const afilename, aitem: string);
var idx : longint;
begin
  if (length(aitem) > 0) then begin
    idx:= ParamsCombo.Items.IndexOf(aitem);
    if (idx <> 0) then begin
      if (idx > 0) then ParamsCombo.Items.Delete(idx);
      ParamsCombo.Items.Insert(0, aitem);
    end;
  end;
  ParamsCombo.Items.SaveToFile(afilename);
end;

procedure TConnStringDialog.SetOpenParams(const adefault: string; aenabled: boolean);
const colors : array[boolean] of tColor = (clBtnface, clWindow);
begin
  OpenEdit.Text:= adefault; OpenEdit.Enabled:= aenabled; OpenLbl.Enabled:= aenabled;
  OpenEdit.Color:= colors[aenabled];
end;

{ misc. routines }

function GetStdDialogParams(const acfgname, acaption, acheckcaption: string;
                            aenableopen, aenablecancel, aenableckeck: boolean;
                            var aopenparams: string; var cancelled, ischecked: boolean): string;
begin
  cancelled:= false;
  setlength(result, 0);
  with TConnStringDialog.Create(nil) do try
    Caption:= acaption;
    CancelBtn.Enabled:= aenablecancel;
    FeatureCheck.Visible:= aenableckeck;
    FeatureCheck.Checked:= ischecked;
    LoadItems(acfgname);
    if not aenableopen then setlength(aopenparams, 0);
    if aenableckeck then FeatureCheck.Caption:= acheckcaption;
    SetOpenParams(aopenparams, aenableopen);
    if (ShowModal = mrOK) then begin
      result:= ParamsCombo.Text;
      if (length(result) > 0) then begin
        SaveItems(acfgname, result);
        if aenableopen then aopenparams:= OpenEdit.Text;
        if aenableckeck then ischecked:= FeatureCheck.Checked;
      end;
    end else cancelled:= true;
  finally free; end;
end;

end.
