{$i replspydefs.pas}

unit replspymain;

interface

uses
  Classes, SysUtils, Forms, StdCtrls, Controls,
  ComCtrls, ExtCtrls, Menus,
  {$ifdef with_production_key} replspyprod, {$endif}
  clipbrd,
  replspytable, replspyconndialog, replspyutils,
  {$ifdef with_cgate_ex} cgate_ex, {$endif}
  gateobjects, cgate;

type
  TReplSpyForm      = class;

  TGuiListener      = class(tCGateListener)
  private
    FNode           : TTreeNode;
    FTree           : TTreeView;
    FForm           : TForm;
    FIsFirstMessage : boolean;
    FTempData       : pChar;
    FTempSize       : longint;
    procedure   LSN_OPEN(var message); message CG_MSG_OPEN;
    procedure   LSN_CLOSE(var message); message CG_MSG_CLOSE;

    procedure   TN_BEGIN(var message); message CG_MSG_TN_BEGIN;

    procedure   MSG_DATA(var message); message CG_MSG_DATA;
    procedure   P2MQ_TIMEOUT(var message); message CG_MSG_P2MQ_TIMEOUT;

    procedure   STREAM_DATA(var message); message CG_MSG_STREAM_DATA;
    procedure   P2REPL_LIFENUM (var message); message CG_MSG_P2REPL_LIFENUM;
    procedure   P2REPL_CLEARDELETED (var message); message CG_MSG_P2REPL_CLEARDELETED;
    procedure   P2REPL_REPLSTATE(var message); message CG_MSG_P2REPL_REPLSTATE;
  protected
    function    updatetablerowcount(atableform: TTableForm): boolean;
  public
    constructor create(AOwner: tCGateConnection; const aparams, aopenparams: string;
                       aNode: TTreeNode; aTree: TTreeView; aForm: TForm); reintroduce; virtual;
    destructor  destroy; override;
    procedure   doParseData(msg_index: longint; adata: pChar; adatasize: longint; anulls: pChar; arev: int64);
    procedure   idle; override;
    procedure   log(const atemplate: string; const aparams: array of const);
    procedure   setguiobjects(aNode: TTreeNode; aTree: TTreeView; aForm: TForm);
  end;

  TGuiPublisher     = class(tCGatePublisher)
  private
    FNode           : TTreeNode;
    FTree           : TTreeView;
    FForm           : TForm;
    FListener       : TGuiListener;
    FBeginTran      : pcg_msg;
    FCommitTran     : pcg_msg;

    function        getBeginTranMsg: pcg_msg;
    function        getCommitTranMsg: pcg_msg;
  public
    constructor create(AOwner: tCGateConnection; const aparams, aopenparams: string;
                       aNode: TTreeNode; aTree: TTreeView; aForm: TForm; aNeedListener: boolean); reintroduce; virtual;
    destructor  destroy; override;
    function    open(const aopenparams: string): longint; override;
    function    close: longint; override;
    function    process(atimeout: longint; var stage: tProcessStage): longint; override;

    property    begin_tran_msg: pcg_msg read getBeginTranMsg;
    property    commit_tran_msg: pcg_msg read getCommitTranMsg;
  end;

  { TReplSpyForm }

  TReplSpyForm = class(TForm)
    MainPnl: TPanel;
    loglist: TListBox;
    ObjectTree: TTreeView;
    Splitter1: TSplitter;
    ObjTreePopup: TPopupMenu;
    OpenTableMenuItem: TMenuItem;
    BottomPnl: TPanel;
    StatusBar1: TStatusBar;
    MainToolBar: TToolBar;
    AddStreamBtn: TToolButton;
    BigCmd: TImageList;
    BigCmdGray: TImageList;
    TreeImages: TImageList;
    ConnectBtn: TToolButton;
    sep1: TToolButton;
    OpenTableBtn: TToolButton;
    DelBtn: TToolButton;
    sep2: TToolButton;
    DeleteMenuItem: TMenuItem;
    AddSvcBtn: TToolButton;
    SendTrsBtn: TToolButton;
    ProcessTimer: TTimer;
    Sep1MenuItem: TMenuItem;
    SendMenuItem: TMenuItem;
    LogPopup: TPopupMenu;
    ClearLogMenuItem: TMenuItem;
    CopyLogMenuItem: TMenuItem;
    CopylineLogMenuItem: TMenuItem;
    procedure doConnectDisconnect(Sender: TObject);
    procedure doProcessConnection(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure AddStreamBtnClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure ObjectTreeDeletion(Sender: TObject; Node: TTreeNode);
    procedure ObjectTreeContextPopup(Sender: TObject; MousePos: TPoint; var Handled: Boolean);
    procedure PopupMenuItemClick(Sender: TObject);
    procedure ObjectTreeChange(Sender: TObject; Node: TTreeNode);
    procedure OnToolBtnClick(Sender: TObject);
    procedure ObjectTreeExit(Sender: TObject);
    procedure ObjectTreeEnter(Sender: TObject);
    procedure AddSvcBtnClick(Sender: TObject);
    procedure ClearLogMenuItemClick(Sender: TObject);
    procedure CopyLogMenuItemClick(Sender: TObject);
    procedure CopylineLogMenuItemClick(Sender: TObject);
  private
    FConnection: tCGateConnection;
    procedure ProcessClickEvents(Sender: TObject; aNode: tTreeNode);
  public
    function  BuildSchemeTree(aobject: tCGateCustom; aTree: TTreeView; aNode: TTreeNode): boolean;

    procedure log(const aevent: string); overload;
    procedure log(const atemplate: string; const aparams: array of const); overload;
  end;

var
  ReplSpyForm: TReplSpyForm;

implementation

{$R *.dfm}

{ TGuiListener }

constructor TGuiListener.create(AOwner: tCGateConnection; const aparams, aopenparams: string;
                               aNode: TTreeNode; aTree: TTreeView; aForm: TForm);
begin
  inherited create(AOwner, aparams);
  openparams:= aopenparams;

  FNode:= anode;
  FTree:= atree;
  FForm:= aForm;

  FTempData:= nil;
  FTempSize:= 0;
end;

destructor TGuiListener.destroy;
begin
  if assigned(FTempData) then freemem(FTempData);
  inherited destroy;
end;

procedure TGuiListener.log(const atemplate: string; const aparams: array of const);
begin if assigned(FForm) and (FForm is TReplSpyForm) then TReplSpyForm(FForm).log(atemplate, aparams); end;

function TGuiListener.updatetablerowcount(atableform: TTableForm): boolean;
begin
  result:= false;
  if assigned(atableform) then with atableform do begin
    if (LV.Items.Count <> Data.Count) then begin
      LV.Items.Count:= Data.Count;
      Caption:= format(CaptionTemplate, [Data.Count]);
      TreeNode.Text:= format('%s [%d]', [TableName, Data.Count]);
      result:= true;
    end;
  end;
end;

procedure TGuiListener.doParseData(msg_index: longint; adata: pChar; adatasize: longint; anulls: pChar; arev: int64);
var frm      : TTableForm;
    addflag  : boolean;
begin
  if assigned(adata) and (msg_index >= 0) and assigned(FNode) then begin
    if (msg_index >= 0) and (msg_index < FNode.Count) then begin
      frm:= TTableForm(FNode.Items[msg_index].Data);
      if assigned(frm) then begin
        with frm.Data do begin
          addflag:= true;

          // micex tables support
          if FIsFirstMessage and MCXClearOnUpd then clear else
          if MCXOrdBook then begin
            if not assigned(FTempData) or (FTempSize < adatasize) then begin
              ReallocMem(FTempData, adatasize);
              FTempSize:= adatasize;
            end;

            if FIsFirstMessage then fillchar(FTempData^, adatasize, 0);

            PrimaryKey.comparefieldcount:= 2;
            try
              if (compare(FTempData, adata) <> 0) then begin
                addflag:= (cleanupbydata(adata, adatasize, anulls) > 2);
                system.move(adata^, FTempData^, adatasize);
              end;
            finally PrimaryKey.comparefieldcount:= -1; end;
          end;

          if addflag then addreplacedata(adata, adatasize, anulls, arev);
        end;
      end;
    end else log('ERROR: Unknown message index: %d', [msg_index]);
  end;
end;

procedure TGuiListener.LSN_OPEN(var message);
const success: array[boolean] of string = ('FAILED', 'OK');
begin
  if assigned(FForm) then log('Listener: %s opened; getting scheme: %s',
                              [params, success[TReplSpyForm(FForm).BuildSchemeTree(Self, FTree, FNode)]]);
end;

procedure TGuiListener.LSN_CLOSE(var message);
begin log('Listener: %s closed.', [params]); end;

procedure TGuiListener.TN_BEGIN(var message);
begin FIsFirstMessage:= true; end;

procedure TGuiListener.P2REPL_LIFENUM(var message);
var i   : longint;
    frm : TTableForm;
begin
  log('Listener: %s LifeNumber changed to: %d', [params, plongint(tcg_msg(message).data)^]);
  if assigned(FNode) then begin
    for i:= 0 to FNode.Count - 1 do begin
      frm:= TTableForm(FNode.Items[i].Data);
      if assigned(frm) then begin
        frm.Data.Clear;
        updatetablerowcount(frm);
      end;
    end;
  end;
end;

procedure TGuiListener.P2REPL_CLEARDELETED(var message);
var  frm : TTableForm;
     cnt : longint;
begin
  with pcg_data_cleardeleted(tcg_msg(message).data)^ do
    if (size_t(table_idx) <> CG_INDEX_INVALID) and assigned(FNode) then begin
      frm:= TTableForm(FNode.Items[table_idx].Data);
      if assigned(frm) then begin
        cnt:= frm.Data.cleanupbyrev(table_rev);
        if (cnt > 0) then begin
          updatetablerowcount(frm);
          log('Listener: %s ClearDeleted: %s rev: %d; deleted %d records', [params, frm.TableName, table_rev, cnt]);
        end;
      end;
    end;
end;

procedure TGuiListener.STREAM_DATA(var message);
begin
  with tcg_msg_streamdata(message) do
    if (msg_index <> CG_INDEX_INVALID) then doParseData(msg_index, data, data_size, nulls, rev);
  FIsFirstMessage:= false;
end;

procedure TGuiListener.MSG_DATA(var message);
begin
  with tcg_msg_data(message) do
    if (msg_index <> CG_INDEX_INVALID) then doParseData(msg_index, data, data_size, nil, 0);
end;

procedure TGuiListener.P2MQ_TIMEOUT(var message);
begin log('Listener: %s; reply: TIMEOUT', [params]); end;

procedure TGuiListener.idle;
var i    : longint;
    frm  : TTableForm;
begin
  if assigned(FNode) then
    for i:= 0 to FNode.Count - 1 do begin
      frm:= TTableForm(FNode.Items[i].Data);
      with frm do begin
        if not updatetablerowcount(frm) and data.NeedRepaint then LV.Invalidate;
        data.NeedRepaint:= false;
      end;
    end;
end;

procedure TGuiListener.setguiobjects(aNode: TTreeNode; aTree: TTreeView; aForm: TForm);
begin FNode:= anode; FTree:= atree; FForm:= aForm; end;

procedure TGuiListener.P2REPL_REPLSTATE(var message);
begin
  with tcg_msg(message) do log('Listener: %s; replstate=%s', [params, pChar(data)]);
end;

{ TGuiPublisher }

constructor TGuiPublisher.create(AOwner: tCGateConnection; const aparams, aopenparams: string;
                                 aNode: TTreeNode; aTree: TTreeView; aForm: TForm; aNeedListener: boolean);
const objid : longint = 0;
begin
  if aNeedListener then begin
    inc(objid);
    inherited create(AOwner, format('%s;name=Object%d', [aparams, objid]));
    FListener:= TGuiListener.create(AOwner,
                                    format('%sreply://;ref=Object%d;%s',
                                           [protocol, objid, FilterParameters(aparams,
                                                                              ['format', 'pricefieldtype', 'category'])
                                           ]),
                                    '', anode, atree, aform);
  end else begin
    inherited create(AOwner, aparams);
    FListener:= nil;
  end;
  openparams:= aopenparams;
  FNode:= anode; FTree:= atree; FForm:= aForm;
  FBeginTran:= nil; FCommitTran:= nil; 
end;

destructor TGuiPublisher.destroy;
begin
  if assigned(FBeginTran) then cg_pub_msgfree(Handle, FBeginTran);
  if assigned(FCommitTran) then cg_pub_msgfree(Handle, FCommitTran);
  if assigned(FListener) then freeandnil(FListener);
  inherited destroy;
end;

function TGuiPublisher.getBeginTranMsg: pcg_msg;
begin
  {$ifdef with_cgate_ex} if not assigned(FBeginTran) then cg_pub_msgalloc(Handle, CG_MSG_TN_BEGIN, 0, FBeginTran); {$endif}
  result:= FBeginTran;
end;

function TGuiPublisher.getCommitTranMsg: pcg_msg;
begin
  {$ifdef with_cgate_ex} if not assigned(FCommitTran) then cg_pub_msgalloc(Handle, CG_MSG_TN_COMMIT, 0, FCommitTran); {$endif}
  result:= FCommitTran;
end;

function TGuiPublisher.open(const aopenparams: string): longint;
begin
  result:= inherited open(aopenparams);
  if (result = CG_ERR_OK) and not assigned(FListener) and assigned(FForm) then TReplSpyForm(FForm).BuildSchemeTree(Self, FTree, FNode);
  if assigned(FForm) then TReplSpyForm(FForm).log('Publisher: %s opened; result: %.8x', [params, result]);
end;

function TGuiPublisher.close: longint;
begin
  result:= inherited close;
  if assigned(FForm) then TReplSpyForm(FForm).log('Publisher: %s closed.', [params]);
end;

function TGuiPublisher.process(atimeout: longint; var stage: tProcessStage): longint;
var tmp : tProcessStage;
begin
  result:= inherited process(atimeout, stage);
  if assigned(FListener) then FListener.process(atimeout, tmp);
end;

{ TReplSpyForm }

procedure TReplSpyForm.FormCreate(Sender: TObject);
var tmp       : string;
    cancelled, checked : boolean;
begin
  EnvironmentSettings:= FilterParameters(GetStdDialogParams('replspy_environment.cfg', 'Set environment parameters', '',
                                                            false, true, false, tmp, cancelled, checked), ['key'], true);
  if cancelled then Application.Terminate;
end;

procedure TReplSpyForm.FormDestroy(Sender: TObject);
begin
  ObjectTree.Items.Clear;
  if assigned(FConnection) then freeandnil(FConnection);
end;

procedure TReplSpyForm.log(const aevent: string);
const max_logstr_len = 1024;
var   ts, tmp        : string;
      idx            : longint;
begin
  with loglist.Items do begin
    ts:= formatdatetime('HH:NN:SS.ZZZ ', now);
    idx:= 1;
    repeat
      tmp:= copy(aevent, idx, max_logstr_len);
      if (length(tmp) > 0) then Add(ts + tmp);
      inc(idx, max_logstr_len);
    until (length(tmp) < max_logstr_len);
    while (Count > 500) do Delete(0);
  end;
end;

procedure TReplSpyForm.log(const atemplate: string; const aparams: array of const);
begin log(format(atemplate, aparams)); end;

procedure TReplSpyForm.ObjectTreeDeletion(Sender: TObject; Node: TTreeNode);
begin
  if assigned(Node) and assigned(Node.data) and (tObject(Node.data) is TObject) then begin
    TObject(Node.data).free;
    Node.data:= nil;
  end;
end;

procedure TReplSpyForm.doConnectDisconnect(Sender: TObject);
const captions           : array[boolean] of string = ('Connect', 'Disconnect');
      imgidxs            : array[boolean] of longint = (0, 1);
var   params, openparams : string;
      cancelled, checked : boolean;
begin
  if not assigned(FConnection) then begin
    openparams:= '';
    params:= GetStdDialogParams('replspy_connections.cfg', 'Set connection parameters', '', false, true, false, openparams, cancelled, checked);
    if not cancelled and (length(params) > 0) then begin
      FConnection:= tCGateConnection.create(params);
      Self.Caption:= format('ReplSpy: %s', [params]);;
      log('Connect: %s', [params]);
    end;
  end else begin
    ObjectTree.Items.Clear;
    log('Disconnect: %s', [FConnection.params]);
    FreeAndNil(FConnection);
    Self.Caption:='ReplSpy';
  end;

  AddStreamBtn.Enabled  := assigned(FConnection);
  AddSvcBtn.Enabled     := assigned(FConnection);
  DelBtn.Enabled        := false;

  OpenTableBtn.Enabled  := false;
  SendTrsBtn.Enabled    := false;

  ConnectBtn.Caption    := captions[assigned(FConnection)];
  ConnectBtn.ImageIndex := imgidxs[assigned(FConnection)];
end;

procedure TReplSpyForm.AddStreamBtnClick(Sender: TObject);
const cfg_name = 'replspy_streams.cfg';
var   itm                : tTreeNode;
      params, openparams : string;
      cancelled, checked : boolean;
begin
  if assigned(FConnection) then begin
    setlength(openparams, 0);
    params:= GetStdDialogParams('replspy_streams.cfg', 'Set stream parameters', '', true, true, false, openparams, cancelled, checked);
    if not cancelled and (length(params) > 0) then begin
      itm:= ObjectTree.Items.AddChild(nil, params);
      itm.Data:= TGuiListener.create(FConnection, params, openparams, itm, ObjectTree, Self);
      itm.ImageIndex:= 1; itm.SelectedIndex:= 1;
    end;
  end;
end;

procedure TReplSpyForm.AddSvcBtnClick(Sender: TObject);
var   itm                : tTreeNode;
      params, openparams : string;
      cancelled, needlsn : boolean;
begin
  if assigned(FConnection) then begin
    setlength(openparams, 0);
    needlsn:= true;

    params:= GetStdDialogParams('replspy_publishers.cfg', 'Set service parameters', 'Reply listener',
                                true, true, true,
                                openparams, cancelled, needlsn);

    if not cancelled and (length(params) > 0) then begin
      itm:= ObjectTree.Items.AddChild(nil, params);
      itm.Data:= TGuiPublisher.create(FConnection, params, openparams, itm, ObjectTree, Self, needlsn);
      itm.ImageIndex:= 13; itm.SelectedIndex:= 13;
    end;
  end;
end;

procedure TReplSpyForm.ObjectTreeContextPopup(Sender: TObject; MousePos: TPoint; var Handled: Boolean);
var node: TTReeNode;
  procedure doPopupMenu(anode: tTreeNode; const aenable, adisable: array of tMenuItem);
  var i : longint;
  begin
    with TTreeView(Sender) do begin
      Selected:= anode;
      MousePos:= ObjectTree.ClientToScreen(MousePos);
      for i:= 0 to length(aenable) - 1 do with aenable[i] do begin
        Tag:= longint(anode);
        Enabled:= true;
      end;
      for i:= 0 to length(adisable) - 1 do adisable[i].Enabled:= false;
      PopupMenu.Popup(MousePos.X, MousePos.Y);
    end;
  end;
begin
  if assigned(Sender) and (Sender is TTreeView) then with TTreeView(Sender) do begin
    ObjectTreeChange(Sender, TTreeView(Sender).Selected);
    node:= GetNodeAt(MousePos.X, MousePos.Y);
    if assigned(node) and assigned(node.data) then begin
      if (tObject(node.data) is TTableForm) then begin
        if TTableForm(node.data).AllowSend then doPopupMenu(node, [OpenTableMenuItem, SendMenuItem], [DeleteMenuItem])
                                           else doPopupMenu(node, [OpenTableMenuItem], [DeleteMenuItem, SendMenuItem]);
      end else
      if (tObject(node.data) is TGuiListener) or (tObject(node.data) is TGuiPublisher) then begin
        doPopupMenu(node, [DeleteMenuItem], [OpenTableMenuItem, SendMenuItem]);
      end;  
    end;
  end;
  Handled:= true;
end;

procedure TReplSpyForm.ObjectTreeChange(Sender: TObject; Node: TTreeNode);
begin
  if assigned(Sender) and (Sender is TTreeView) then
    if assigned(TTreeView(Sender).Selected) then begin
      with TTreeView(Sender).Selected do begin
        OpenTableBtn.Enabled:= assigned(data) and (tObject(data) is TTableForm);
        SendTrsBtn.Enabled:= assigned(data) and (tObject(data) is TTableForm) and TTableForm(data).AllowSend;
        DelBtn.Enabled:= assigned(data) and ((tObject(data) is TGuiListener) or (tObject(data) is TGuiPublisher));
      end;
    end else DelBtn.Enabled:= false;
end;

procedure TReplSpyForm.ProcessClickEvents(Sender: TObject; aNode: tTreeNode);
var node  : tTreeNode;
begin
  if assigned(aNode.Data) then
    if (tObject(aNode.Data) is TTableForm) then begin
      if ((Sender = OpenTableMenuItem) or (Sender = OpenTableBtn)) then TTableForm(aNode.Data).Show else
      if ((Sender = SendMenuItem) or (Sender = SendTrsBtn)) and TTableForm(aNode.Data).AllowSend then begin
        node:= aNode.Parent;
        if assigned(node) and assigned(node.Data) and (tObject(node.Data) is TGuiPublisher) then
          TTableForm(aNode.Data).SendTransaction(TGuiPublisher(node.Data));
      end;
    end else
    if ((tObject(aNode.Data) is TGuiListener) or (tObject(aNode.Data) is TGuiPublisher)) then begin
      ObjectTree.Items.Delete(aNode);
      ObjectTreeChange(ObjectTree, nil);
    end;
end;

procedure TReplSpyForm.PopupMenuItemClick(Sender: TObject);
begin
  if assigned(Sender) and (Sender is TMenuItem) and (TMenuItem(Sender).Tag <> 0) then
    if tObject(TMenuItem(Sender).Tag) is TTreeNode then ProcessClickEvents(Sender, TTreeNode(TMenuItem(Sender).Tag));
end;

procedure TReplSpyForm.OnToolBtnClick(Sender: TObject);
begin
  if assigned(Sender) and assigned(ObjectTree.Selected) then ProcessClickEvents(Sender, ObjectTree.Selected);
end;

procedure TReplSpyForm.ObjectTreeEnter(Sender: TObject);
begin ObjectTreeChange(Sender, nil); end;

procedure TReplSpyForm.ObjectTreeExit(Sender: TObject);
begin
  OpenTableBtn.Enabled:= false;
  SendTrsBtn.Enabled:= false;
  DelBtn.Enabled:= false;
end;

procedure TReplSpyForm.doProcessConnection(Sender: TObject);
const max_process_messages = 50;
var   res            : longint;
      itm            : tTreeNode;
      processed_msgs : longint;
      stage          : tProcessStage;
begin
  if assigned(FConnection) then begin
    processed_msgs:= 0;
    repeat
      res:= FConnection.process(0, stage);
      inc(processed_msgs);
    until (res <> CG_ERR_OK) or (processed_msgs >= max_process_messages);
    if (stage = psOpen) and (res <> CG_ERR_OK) then begin
      doConnectDisconnect(nil);
      Application.MessageBox(pChar(format('Error (re)opening connection: %.8x (%d)', [res, res])), 'Error', 0);
      ProcessTimer.Interval:= 250;
    end else begin
      if (res = CG_ERR_OK) or (res = CG_ERR_TIMEOUT) then begin
        itm:= ObjectTree.Items.GetFirstNode;
        while assigned(itm) do try
          if assigned(itm.data) and (tObject(itm.data) is TCGateCustom) then TCGateCustom(itm.data).process(0, stage);
        finally itm:= itm.GetNextChild(itm); end;
      end;
      ProcessTimer.Interval:= 10;
    end;
  end else ProcessTimer.Interval:= 250;
end;

procedure TReplSpyForm.ClearLogMenuItemClick(Sender: TObject);
begin loglist.Items.Clear; end;

procedure TReplSpyForm.CopyLogMenuItemClick(Sender: TObject);
var tmp : string;
begin
  tmp:= loglist.Items.Text;
  if (length(tmp) > 0) then Clipboard.SetTextBuf(pChar(tmp));
end;

procedure TReplSpyForm.CopylineLogMenuItemClick(Sender: TObject);
var tmp : string;
begin
  if loglist.ItemIndex > 0 then begin
    tmp:= loglist.Items[loglist.ItemIndex];
    if (length(tmp) > 0) then Clipboard.SetTextBuf(pChar(tmp));
  end;
end;

function TReplSpyForm.BuildSchemeTree(aobject: tCGateCustom; aTree: TTreeView; aNode: TTreeNode): boolean;
var   scheme        : pcg_scheme_desc;
      msgs          : pcg_message_desc;
      flds          : pcg_field_desc;
      idxdef        : pcg_index_desc;
      idxflds       : pcg_indexfield_desc;
      itm           : TTreeNode;
      tmpn1, tmpn2  : TTreeNode;
      frm           : TTableForm;
      extr          : TFieldExtractor;
      idx, iconidx  : longint;
      i, j          : longint;
      tmp           : string;
      sl, tmpsl     : tStringList;
const imgidx        : array[boolean, 0..2] of longint = ((2, 2, 2), (20, 6, 7));
      algnmt        : array[boolean] of tAlignment = (taLeftJustify, taRightJustify);
      ascdesc       : array[boolean] of longint = (22, 23);
  function seticoidx(anode: tTreeNode; aidx: longint): tTreeNode;
  begin result:= anode; if assigned(result) then with result do begin ImageIndex:= aidx; SelectedIndex:= aidx; end; end;
begin
  result:= false;
  if assigned(aobject) and assigned(ANode) and assigned(ATree) then try
    if ANode.HasChildren then ANode.DeleteChildren;
    scheme:= aobject.getscheme;
    if assigned(scheme) then begin
      msgs:= scheme^.messages;
      while assigned(msgs) do try

        frm:= TTableForm.Create(Self);
        with frm do begin
          TableID         := msgs^.id;
          TableName       := msgs^.name;
          CaptionTemplate := format('%s [%%d]', [msgs^.name]);
          Caption         := format(frm.CaptionTemplate, [0]);
          AllowSend       := (tObject(ANode.Data) is TGuiPublisher);
          MainForm        := Self;
          if aobject is TCGateListener then Listener:= TCGateListener(aobject) else Listener:= nil;
          TreeNode        := ATree.Items.AddChildObject(ANode, msgs^.name, frm);

          itm:= TreeNode;

          Data.Extractors.Clear;
          LV.Columns.Add.Width:= 0; // dummy column

          idx:= 0;
          flds:= msgs^.fields;
          while assigned(flds) do try

            extr:= TFieldExtractor.Create(flds^, idx);
            Data.Extractors.Add(extr);

            // plaza2 replication fields support
            if (comparetext(flds^.name, 'replRev') = 0) then begin
              Data.RevisionKey.clear;
              Data.RevisionKey.addkeyfield(flds^.field_type, flds^.size, flds^.offset);
            end else
            if (comparetext(flds^.name, 'replAct') = 0) then Data.ReplActOffset:= flds^.offset;

            with LV.Columns.Add do begin
              if assigned(flds^.desc) then Caption:= flds^.desc else Caption:= flds^.name;
              Tag       := longint(extr);
              Alignment := algnmt[(upcase(extr.FieldType[1]) in ['I', 'U', 'D'])];
              Width     := 70;
            end;

            seticoidx(ATree.Items.AddChild(itm, format('%s [%s] %s', [string(flds^.name), string(flds^.field_type), string(flds^.desc)])), 9);
          finally
            flds:= flds^.next;
            inc(idx);
          end;

          Data.FieldCount   := idx;
          Data.RowSize      := msgs^.size;

          // msgid support
          if (msgs^.id <> 0) then seticoidx(ATree.Items.AddChild(itm, format('MSGID=%d', [msgs^.id])), 3);

          iconidx:= 0;

          // hints support
          if assigned(msgs^.hints) then begin
            Data.Hints:= string(msgs^.hints);

            sl:= tStringList.Create;
            try
              DecodeCommaText(string(msgs^.hints), sl, ';');
              if (length(sl.Values['request']) > 0) then iconidx:= 1 else
              if (length(sl.Values['reply']) > 0)   then iconidx:= 2;

              tmpn1:= seticoidx(ATree.Items.AddChild(itm, 'HINTS'), 10);
              for i:= 0 to sl.Count - 1 do begin
                tmp:= sl[i];
                if (pos('=', tmp) > 0) then begin
                  tmpn2:= seticoidx(ATree.Items.AddChild(tmpn1, sl.Names[i]), 10);
                  tmp:= sl.Values[sl.Names[i]];
                  if (length(tmp) > 0) then begin
                    tmpsl:= tStringList.Create;
                    try
                      if (pos(',', tmp) > 0) then DecodeCommaText(tmp, tmpsl, ',') else
                      if (pos('|', tmp) > 0) then DecodeCommaText(tmp, tmpsl, '|');
                      if (tmpsl.count > 0) then begin
                        for j:= 0 to tmpsl.count - 1 do seticoidx(ATree.Items.AddChild(tmpn2, tmpsl[j]), 10);
                      end else seticoidx(ATree.Items.AddChild(tmpn2, tmp), 10);
                    finally tmpsl.Free; end;
                  end;
                end else seticoidx(ATree.Items.AddChild(tmpn1, tmp), 10);
              end;
            finally sl.free; end;
          end;

          seticoidx(itm, imgidx[AllowSend, iconidx]);

          // index support
          idx:= 0;
          idxdef:= msgs^.indices;
          if assigned(idxdef) then begin
            tmpn1:= seticoidx(ATree.Items.AddChild(itm, 'INDEXES'), 21);
            while assigned(idxdef) do try
              tmpn2:= seticoidx(ATree.Items.AddChild(tmpn1, string(idxdef^.name)), 21);
              idxflds:= idxdef^.fields;
              if assigned(idxflds) and (idx = 0) then Data.PrimaryKey.clear; // index number 0 is always primary
              while assigned(idxflds) do try
                if assigned(idxflds^.field) then with idxflds^.field^ do begin
                  if (idx = 0) then Data.PrimaryKey.addkeyfield(field_type, size, offset); // add fields for primary index
                  seticoidx(ATree.Items.AddChild(tmpn2, string(name)), ascdesc[boolean(idxflds^.sort_order)]);
                end;
              finally idxflds:= idxflds^.next; end;
            finally
              idxdef:= idxdef^.next;
              inc(idx);
            end;
          end;

        end;

      finally msgs:= msgs^.next; end;

      result:= true;
    end;

  finally
    ANode.Expand(false);
    ObjectTreeChange(ATree, nil);
  end;
end;

end.
