{$i replspydefs.pas}

unit replspysendtrsform;

interface

uses
  Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, ExtCtrls, Math, Spin, EditBtn,
  LCLType,
  {$ifdef unix} lconvencoding, {$endif}
  cgate, gateutils;

const control_enabled  = true;
      control_disabled = false;

type

  { TSendTrsForm }

  TSendTrsForm = class(TForm)
    SB: TScrollBox;
    BP: TPanel;
    OKBtn: TButton;
    CancelBtn: TButton;
    DumpCheck: TCheckBox;
    procedure FormCreate(Sender: TObject);
  private
    FItemTop: longint;
  public
    procedure AddField(const aname: string; const atype: string; afldno, asize: longint; adata: pointer = nil);
    procedure SetFieldBuf(var abuf; actl: tControl; const atype: string);

    property  LastItemTop: longint read FItemTop write FItemTop;
  end;

implementation

{$R *.dfm}

{ TSendTrsForm }

procedure TSendTrsForm.FormCreate(Sender: TObject);
begin
  FItemTop:= 5;
end;

procedure TSendTrsForm.AddField(const aname: string; const atype: string; afldno, asize: longint; adata: pointer);
const bounds_min         = false;
      bounds_max         = true;
const spinbounds : array[boolean, 1..8, boolean] of int64 =
        (((low(ShortInt), high(ShortInt)), (low(SmallInt), high(SmallInt)), (0, 0), (low(LongInt), high(LongInt)),
          (0, 0), (0, 0), (0, 0), (low(Int64), high(Int64))),
         ((0, high(Byte)), (0, high(Word)), (0, 0), (0, high(Cardinal)),
          (0, 0), (0, 0), (0, 0), (0, high(Int64))));
var   itm        : tControl;
      FldType    : char;
      tempint    : int64;
      tempdt     : tDateTime;
      scale      : byte;
  function SetControlParams(actl: tControl; aenabled: boolean; aleft, awidth: longint): tControl;
  begin
    result:= actl;
    if assigned(result) then with result do begin
      Parent:= SB; Enabled:= aenabled; tag:= afldno; hint:= aname;
      top:= FItemTop; left:= aleft; width:= awidth;
    end;
  end;
  function MakeNotSupported(AOwner: TComponent; const atype: string): tControl;
  begin
    result:= SetControlParams(TEdit.Create(AOwner), control_disabled, 120, 160);
    tEdit(result).Text:= format('<type %s is not supported>', [atype]);
  end;
  procedure SetMinMax(actl: TSpinEdit; const amin, amax: int64);
  begin actl.MinValue:= amin; actl.MaxValue:= amax; end;
begin
  scale:= 0; tempint:= 0;

  with tLabel.create(SB) do begin
    parent:= SB; autosize:= true; caption:= aname; tag:= -1;
    top:= FItemTop + 2; left:= 5;
  end;

  if (length(atype) > 0) then FldType:= UpCase(atype[1]) else FldType:= #0;

  case FldType of
    'C'      : begin
                 itm:= SetControlParams(TEdit.Create(SB), control_enabled, 120, 160);
                 TEdit(itm).MaxLength:= asize - 1;
                 {$ifdef Unix}
                 if assigned(adata) then TEdit(itm).Text:= UTF8ToCP1251(pChar(adata));
                 {$else}
                 if assigned(adata) then TEdit(itm).Text:= pChar(adata);
                 {$endif}
               end;
    'I', 'U' : if (asize in [1..8]) then begin
                 itm:= SetControlParams(TSpinEdit.Create(SB), control_enabled, 120, 160);
                 SetMinMax(TSpinEdit(itm), spinbounds[(FldType = 'U'), asize, bounds_min], spinbounds[(FldType = 'U'), asize, bounds_max]);
                 if assigned(adata) then TSpinEdit(itm).Value:= cg_utils_get_int(pChar(atype), adata^);
               end else itm:= MakeNotSupported(SB, atype);
    'D'      : begin
                 itm:= SetControlParams(TFloatSpinEdit.Create(SB), control_enabled, 120, 160);
                 if assigned(adata) then begin
                   if (cg_bcd_get(adata, tempint, scale) = CG_ERR_OK) then TFloatSpinEdit(itm).Value:= tempint * intpower(10, -scale);
                 end;
               end;
    'T'      : begin
                 if assigned(adata) then tempdt:= cg_utils_get_time(adata^) else tempdt:= 0.0;
                 itm:= SetControlParams(TEdit.Create(SB), control_enabled, 120, 160);
                 TEdit(itm).MaxLength:= 20;
                 TEdit(itm).Text:= FormatDateTime('DD/MM/YYYY HH:NN:SS', tempdt);
               end;
    else       itm:= MakeNotSupported(SB, atype);
  end;

  if assigned(itm) then inc(FItemTop, itm.height + 2);
end;

procedure TSendTrsForm.SetFieldBuf(var abuf; actl: tControl; const atype: string);
begin
  if (actl.Enabled = control_enabled) and (length(atype) > 0) then try
    if (actl is TSpinEdit) then cg_utils_set_int(pChar(atype), StrToInt64Def(TEdit(actl).Text, 0), abuf) else
    if (actl is TEdit) then begin
      case upcase(atype[1]) of
        'C' : cg_utils_set_char(pChar(atype), TEdit(actl).Text, abuf);
        'T' : cg_utils_set_datetime(pChar(atype), StrToDateTime(TEdit(actl).Text), set_date_time_date or set_date_time_time, abuf)
      end;
    end else
    if (actl is TFloatSpinEdit) then cg_utils_set_bcd(pChar(atype), TFloatSpinEdit(actl).value, abuf);
  except
    on e: exception do Application.MessageBox(pChar(e.message), 'ERROR:', MB_ICONERROR or MB_OK);
  end;
end;

end.
