{$i replspydefs.pas}

unit replspytable;

interface

uses
  SysUtils, Classes, Controls,
  Forms, Dialogs, ComCtrls, Math, Menus,
  LCLType,
  SortedList, clipbrd,
  replspysendtrsform, replspyutils,
  {$ifdef with_cgate_ex} cgate_ex, {$endif}
  {$ifdef unix} lconvencoding, {$endif}
  GateObjects, gateutils, cgate;

type
  TExtractFunc     = function(const abuffer): string of object;

  TFieldExtractor  = class(TObject)
  private
    FID            : longint;
    FNo            : longint;
    FName          : string;
    FFullType      : string;
    FSize          : longint;
    FOffset        : longint;
    FDesc          : string;
  public
    constructor Create(const afielddesc: tcg_field_desc; afieldno: longint);
    function    ExtractField(const abuffer): string;
    function    ExtractFieldFromRow(adataptr, nullsptr: pointer): string;

    property    FieldType: string read FFullType;
    property    FieldSize: longint read FSize;
  end;

  TKeyField        = record
    FieldType      : string;
    FieldSize      : longint;
    FieldOffset    : longint;
  end;

  TPrimaryKey      = class(TObject)
  private
    FKeyFields     : array of TKeyField;
    FCmpFldCount   : longint;
    function    fGetFieldCount: longint;
    function    fGetField(aindex: longint): TKeyField;
  public
    constructor create;
    function    compare(item1, item2: pointer): longint; virtual;
    procedure   addkeyfield(const atype: string; asize, aoffset: longint);
    procedure   clear;
    property    count: longint read fGetFieldCount;
    property    fields[aindex: longint]: TKeyField read fGetField; default;
    property    comparefieldcount: longint read FCmpFldCount write FCmpFldCount;
  end;

  TDataList        = class(TSortedList)
  private
    FExtractors    : TList;
    FNeedRepaint   : boolean;
    FFieldCount    : longint;
    FReplActOffset : longint;
    FRowSize       : longint;
    FPrimaryKey    : TPrimaryKey;
    FRevisionKey   : TPrimaryKey;
    FMCXOrdBook    : boolean;
    FMCXClearOnUpd : boolean;
  protected
    function    AllocRow: pointer;
    function    GetRowDataPtr(aitem: pointer): pointer;
    function    GetNullsPtr(aitem: pointer): pChar;
    function    GetRevPtr(aitem: pointer): pInt64;
    function    GetPrimaryKey: TPrimaryKey;
    function    GetRevisionKey: TPrimaryKey;
    procedure   SetHints(const ahints: string);
  public
    constructor create;
    destructor  destroy; override;
    procedure   freeitem(item:pointer); override;
    function    checkitem(item: pointer): boolean; override;
    function    compare(item1, item2: pointer): longint; override;
    function    addreplacedata(abuffer: pChar; abuffersize: longint; anulls: pChar; arev: int64): longint;
    function    cleanupbyrev(const arev: int64): longint;
    function    cleanupbydata(abuffer: pChar; abuffersize: longint; anulls: pChar): longint;

    property    NeedRepaint: boolean read FNeedRepaint write FNeedRepaint;
    property    FieldCount: longint read FFieldCount write FFieldCount;
    property    Extractors: TList read FExtractors;

    property    PrimaryKey: TPrimaryKey read GetPrimaryKey;
    property    RevisionKey: TPrimaryKey read GetRevisionKey;

    property    RowSize: longint read FRowSize write FRowSize;

    property    ReplActOffset: longint write FReplActOffset;
    property    Hints: string write SetHints;

    property    MCXOrdBook: boolean read FMCXOrdBook;
    property    MCXClearOnUpd: boolean read FMCXClearOnUpd;
  end;

  TTableForm       = class(TForm)
    LV             : TListView;
    ListViewPopup  : TPopupMenu;
    ClipCopyMenutem: TMenuItem;
    SaveTableDialog: TSaveDialog;
    SavetoCSVMenuItem: TMenuItem;
    N1             : TMenuItem;
    ResendMenuItem : TMenuItem;
    SendMenuItem   : TMenuItem;
    procedure LVData(Sender: TObject; Item: TListItem);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure FormDestroy(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure ClipCopyMenutemClick(Sender: TObject);
    procedure SavetoCSVMenuItemClick(Sender: TObject);
    procedure ResendMenuItemClick(Sender: TObject);
    procedure ListViewPopupPopup(Sender: TObject);
    procedure SendMenuItemClick(Sender: TObject);
  private
    procedure CreateCSVBuffer(abuffer: tStream; aseparator: char);
  public
    Data            : TDataList;
    TableID         : longint;
    TableName       : string;
    CaptionTemplate : string;
    AllowSend       : boolean;
    MainForm        : TForm;
    Listener        : TCGateListener;
    TreeNode        : TTreeNode;

    procedure SendTransaction(aPublisher: tCGatePublisher; msgdata: pChar = nil);
    procedure log(const atemplate: string; const aparams: array of const);
  end;

function getuniquereplrev: int64;

implementation

uses replspymain;

{$R *.dfm}

function getuniquereplrev: int64;
const uniquerev: int64 = 0;
begin
  inc(uniquerev);
  result:= uniquerev;
end;

{ TFieldExtractor }

constructor TFieldExtractor.Create(const afielddesc: tcg_field_desc; afieldno: longint);
begin
  inherited Create;
  FFullType:= afielddesc.field_type;
  FId     := afielddesc.id;
  FNo     := afieldno;
  FSize   := afielddesc.size;
  FOffset := afielddesc.offset;
  FName   := afielddesc.name;
  if assigned(afielddesc.desc) then FDesc:= afielddesc.desc else FDesc:= FName;
end;

function TFieldExtractor.ExtractField(const abuffer): string;
var len : cardinal;
    res : longint;
    buf : array[0..1024] of char;
begin
  len:= sizeof(buf) - 1;
  res:= cg_getstr(pChar(FFullType), pChar(@abuffer) + FOffset, @buf, len);
  case res of
    CG_ERR_BUFFERTOOSMALL : begin
                              setlength(result, len - 1);
                              cg_getstr(pChar(FFullType), pChar(@abuffer) + FOffset, @result[1], len);
                            end;
    CG_ERR_OK             : setstring(result, buf, len - 1);
    else                    setlength(result, 0);
  end;
  {$ifdef Unix} result:= CP1251ToUTF8(result); {$endif}
end;

function TFieldExtractor.ExtractFieldFromRow(adataptr, nullsptr: pointer): string;
begin
   if assigned(adataptr) and assigned(nullsptr) then begin
     case pChar(nullsptr)[FNo] of
       #$0 : result:= ExtractField(adataptr^);
       #$1 : result:= '<null>';
       #$2 : result:= '<undefined>';
       else  result:= '???';
     end;
   end else setlength(result, 0);
end;

{ TPrimaryKey }

constructor TPrimaryKey.create;
begin
  inherited create;
  FCmpFldCount:= -1;
end;

function TPrimaryKey.fGetFieldCount: longint;
begin result:= length(FKeyFields); end;

function TPrimaryKey.fGetField(aindex: longint): TKeyField;
begin
  if (aindex >= 0) and (aindex < length(FKeyFields)) then begin
    result:= FKeyFields[aindex];
  end else begin
    with result do begin setlength(FieldType, 0); FieldSize:= 0; FieldOffset:= 0; end;
  end;
end;

function TPrimaryKey.compare(item1, item2: pointer): longint;
var i, cnt : longint;
    v1, v2 : int64;
    scale  : byte;
  function cmpi64(const v1, v2: int64): longint;
  var tmp : int64;
  begin
    tmp:= v1 - v2;
    if (tmp < 0) then result:= -1 else
    if (tmp > 0) then result:= 1 else result:= 0;
  end;
  function cmpdbl(const v1, v2: double): longint;
  var tmp : double;
  begin
    tmp:= v1 - v2;
    if (tmp < 0) then result:= -1 else
    if (tmp > 0) then result:= 1 else result:= 0;
  end;
begin
  result:= 0; scale:= 0; v1:= 0; v2:= 0;
  if (FCmpFldCount < 0) then cnt:= length(FKeyFields)
                        else cnt:= min(length(FKeyFields), FCmpFldCount);
  i:= 0;
  while (i < cnt) do
    with FKeyFields[i] do begin
      case UpCase(FieldType[1]) of
        'I', 'U' : result:= cmpi64(cg_utils_get_int(pChar(FieldType), (pChar(item1) + FieldOffset)^),
                                   cg_utils_get_int(pChar(FieldType), (pChar(item2) + FieldOffset)^));
        'C'      : result:= StrLComp((pChar(item1) + FieldOffset), (pChar(item2) + FieldOffset), FieldSize);
        'T'      : result:= cmpi64(cg_utils_get_time_as_int((pChar(item1) + FieldOffset)^),
                                   cg_utils_get_time_as_int((pChar(item2) + FieldOffset)^));
        'D'      : begin
                     cg_bcd_get(pChar(item1) + FieldOffset, v1, scale);
                     cg_bcd_get(pChar(item2) + FieldOffset, v2, scale);
                     result:= cmpi64(v1, v2);
                   end;
      end;
      if (result = 0) then inc(i) else i:= cnt;
    end;
end;

procedure TPrimaryKey.addkeyfield(const atype: string; asize, aoffset: longint);
var len : longint;
begin
  len:= length(FKeyFields);
  setlength(FKeyFields, len + 1);
  with FKeyFields[len] do begin FieldType:= atype; FieldSize:= asize; FieldOffset:= aoffset; end;
end;

procedure TPrimaryKey.clear;
begin setlength(FKeyFields, 0); end;

{ TDataList }

constructor TDataList.create;
begin
  inherited create;
  fDuplicates    := dupAccept;
  FExtractors    := TList.create;
  FReplActOffset := -1;
  FRowSize       := 0;
  FMCXOrdBook    := false;
  FMCXClearOnUpd := false;
end;

destructor TDataList.destroy;
begin
  if assigned(Extractors) then freeandnil(FExtractors);
  if assigned(FPrimaryKey) then freeandnil(FPrimaryKey);
  if assigned(FRevisionKey) then freeandnil(FRevisionKey);
  inherited destroy;
end;

procedure TDataList.freeitem(item: pointer);
begin if assigned(item) then freemem(item); end;

function TDataList.checkitem(item: pointer): boolean;
begin result:= assigned(item); end;

function TDataList.compare(item1, item2: pointer): longint;
begin
  if assigned(FPrimaryKey) then begin
    result:= FPrimaryKey.compare(GetRowDataPtr(item1), GetRowDataPtr(item2));
  end else result:= 0;
end;

function TDataList.AllocRow: pointer;
var sz : longint;
begin
  sz:= FRowSize + FFieldCount + sizeof(int64);
  result:= allocmem(sz);
end;

function TDataList.GetRowDataPtr(aitem: pointer): pointer;
begin result:= aitem; end;

function TDataList.GetNullsPtr(aitem: pointer): pChar;
begin result:= pchar(aitem) + FRowSize; end;

function TDataList.GetRevPtr(aitem: pointer): pInt64;
begin result:= pInt64(pchar(aitem) + FRowSize + FFieldCount); end;

function TDataList.GetPrimaryKey: TPrimaryKey;
begin
  if not assigned(FPrimaryKey) then FPrimaryKey:= TPrimaryKey.Create;
  result:= FPrimaryKey;
end;

function TDataList.GetRevisionKey: TPrimaryKey;
begin
  if not assigned(FRevisionKey) then FRevisionKey:= TPrimaryKey.Create;
  result:= FRevisionKey;
end;

procedure TDataList.SetHints(const ahints: string);
var sl: tStringList;
begin
  sl:= tStringList.Create;
  try
    DecodeCommaText(ahints, sl, ';');
    FMCXOrdBook    := (sl.IndexOf('ORDERBOOK') >= 0);
    FMCXClearOnUpd := (sl.IndexOf('CLEAR_ON_UPDATE') >= 0);
  finally sl.free; end;
end;

function TDataList.addreplacedata(abuffer: pChar; abuffersize: longint; anulls: pChar; arev: int64): longint;
var   del       : boolean;
      itm       : pChar;
      i         : longint;
      rowdata   : pChar;
      nullsdata : pChar;
begin
  result:= 0;
  if assigned(abuffer) then begin
    itm:= nil;
    del:= (FReplActOffset >= 0) and (pInt64(abuffer + FReplActOffset)^ <> 0);
    if assigned(FPrimaryKey) then begin
      if not search(abuffer, result) then begin
        if not del then begin
          itm:= AllocRow;
          system.move(abuffer^, GetRowDataPtr(itm)^, abuffersize);
          if assigned(anulls) then system.move(anulls^, GetNullsPtr(itm)^, FFieldCount)
                              else fillchar(GetNullsPtr(itm)^, FFieldCount, 0);
          insert(result, itm);
          FNeedRepaint:= true;
        end;
      end else begin
        if not del then begin
          itm:= Items[result];
          rowdata:= GetRowDataPtr(itm);

          if assigned(anulls) then begin
            nullsdata:= GetNullsPtr(itm);

            for i:= 0 to Extractors.Count - 1 do
              with TFieldExtractor(Extractors[i]) do
                case anulls[FNo] of
                  #$0, #$1 : begin
                               system.move((abuffer + FOffset)^, (rowdata + FOffset)^, FSize);
                               nullsdata[FNo]:= anulls[FNo];
                             end;
                  #$2      : ; // undefined
                end;

          end else system.move(abuffer^, rowdata^, abuffersize);
          FNeedRepaint:= true;
        end else begin
          delete(result);
          FNeedRepaint:= true;
        end;
      end;
    end else begin
      if not del then begin
        itm:= AllocRow;
        system.move(abuffer^, GetRowDataPtr(itm)^, abuffersize);
        if assigned(anulls) then system.move(anulls^, GetNullsPtr(itm)^, FFieldCount)
                            else fillchar(GetNullsPtr(itm)^, FFieldCount, 0);
        result:= count;
        insert(result, itm);
        FNeedRepaint:= true;
      end;
    end;
    if not del and assigned(itm) then GetRevPtr(itm)^:= arev;
  end;
end;

function TDataList.cleanupbyrev(const arev: int64): longint;
var i        : longint;
    itemdata : pChar;
begin
  result:= 0;
  if assigned(FRevisionKey) then begin
    itemdata:= allocmem(FRowSize);
    try
      if (FRevisionKey.count > 0) then system.move(arev, (itemdata + FRevisionKey[0].FieldOffset)^, sizeof(arev));
      for i:= count - 1 downto 0 do
        if FRevisionKey.compare(GetRowDataPtr(items[i]), itemdata) < 0 then begin
          delete(i);
          inc(result);
        end;
    finally freemem(itemdata); end;
  end;
end;

function TDataList.cleanupbydata(abuffer: pChar; abuffersize: longint; anulls: pChar): longint;
var idx: longint;
begin
  result:= 0;
  for idx:= 0 to FFieldCount - 1 do if (anulls[idx] = #0) then inc(result);
  if assigned(FPrimaryKey) and search(abuffer, idx) then repeat
    delete(idx);
  until (idx >= count) or (compare(abuffer, items[idx]) <> 0);
end;

{ TTableForm }

procedure TTableForm.FormCreate(Sender: TObject);
begin
  data:= TDataList.Create;
  LV.DoubleBuffered:= true;
end;

procedure TTableForm.FormDestroy(Sender: TObject);
var i : longint;
begin
  LV.Columns.BeginUpdate;
  try
    for i:= LV.Columns.Count - 1 downto 0 do begin
      with LV.Columns[i] do
        if (Tag <> 0) then begin
          TFieldExtractor(Tag).Free;
          Tag:= 0;
        end;
    end;
  finally LV.Columns.EndUpdate; end;
  if assigned(data) then freeandnil(data);
end;

procedure TTableForm.LVData(Sender: TObject; Item: TListItem);
var i       : longint;
    rowdata : pointer;
    rev     : int64;
begin
  if (Item.Index < Data.Count) then begin
    rowdata:= Data[Item.Index];
    if assigned(rowdata) then begin
      rev:= Data.GetRevPtr(rowdata)^;
      if (rev <> 0) then Item.Caption:= IntToStr(rev) else Item.Caption:= '';
      for i:= 0 to Data.Extractors.Count - 1 do
        Item.SubItems.Add(TFieldExtractor(Data.Extractors[i]).ExtractFieldFromRow(rowdata, Data.GetNullsPtr(rowdata)));
    end;
  end;
end;

procedure TTableForm.log(const atemplate: string; const aparams: array of const);
begin if assigned(MainForm) and (MainForm is TReplSpyForm) then TReplSpyForm(MainForm).log(atemplate, aparams); end;

procedure TTableForm.FormClose(Sender: TObject; var Action: TCloseAction);
begin Action:= caHide; end;

procedure TTableForm.CreateCSVBuffer(abuffer: tStream; aseparator: char);
const crlf    : pchar = #$0d#$0a;
      nul     : char  = #0;
var   i, j, k : longint;
      tmp     : array of TFieldExtractor;
      temp    : string;
      rowdata : pointer;
begin
  if assigned(abuffer) then begin
    setlength(tmp, LV.Columns.Count);
    k:= 0;
    for i:= 0 to LV.Columns.Count - 1 do begin
      tmp[i]:= TFieldExtractor(LV.Columns[i].Tag);
      if (k > 0) then abuffer.write(aseparator, sizeof(aseparator));
      if assigned(tmp[i]) then begin
        temp:= tmp[i].FName;
        abuffer.write(temp[1], length(temp));
        inc(k);
      end;
    end;
    abuffer.write(crlf^, 2);
    for i:= 0 to Data.Count - 1 do begin
      k:= 0;
      rowdata:= Data[i];
      if assigned(rowdata) then
        for j:= 0 to length(tmp) - 1 do begin
          if (k > 0) then abuffer.write(aseparator, sizeof(aseparator));
          if assigned(tmp[j]) then begin
            temp:= tmp[j].ExtractFieldFromRow(rowdata, Data.GetNullsPtr(rowdata));
            abuffer.write(temp[1], length(temp));
            inc(k);
          end;
        end;
      abuffer.write(crlf^, 2);
    end;
    abuffer.write(nul, 1);
  end;
end;

procedure TTableForm.ClipCopyMenutemClick(Sender: TObject);
var ms : tMemoryStream;
begin
  ms:= tMemoryStream.Create;
  with ms do try
    CreateCSVBuffer(ms, #$9);
    Clipboard.SetTextBuf(Memory);
  finally free; end;
end;

procedure TTableForm.SavetoCSVMenuItemClick(Sender: TObject);
var strm : tFileStream;
begin
  with SaveTableDialog do begin
    DefaultExt:= '.csv';
    FileName:= format('%s.csv', [TableName]);
    if Execute then try
      strm:= tFileStream.Create(FileName, fmCreate);
      try CreateCSVBuffer(strm, ';');
      finally strm.free; end;
    except on e: exception do Application.MessageBox(pChar(format('Error writing to file: %s', [e.message])), 'Error', 0); end;
  end;
end;

procedure TTableForm.SendTransaction(aPublisher: tCGatePublisher; msgdata: pChar);
const need_reply : array[boolean] of longint = (0, CG_PUB_NEEDREPLY);
const logfmtstr  = 'Publisher: %s; result: 0x%.8x';
var   i, fldno   : longint;
      msg        : pcg_msg;
      ctl        : tControl;
      flddata    : pChar;
      {$ifdef with_cgate_ex}
      in_trs     : boolean;
      {$endif}
begin
  if assigned(aPublisher) then begin
    with TSendTrsForm.Create(nil) do try
      Caption:= format('Send transaction: %s', [TableName]);
      ClientWidth:= 305;

      msg:= aPublisher.msg_new(CG_KEY_NAME, pChar(TableName));
      if assigned(msg) and assigned(msg.data) then try
        for i:= 0 to Data.Extractors.Count - 1 do
          with TFieldExtractor(Data.Extractors[i]) do begin
            flddata:= msgdata;
            if assigned(flddata) then inc(flddata, FOffset);
            AddField(FDesc, FFullType, i, FSize, flddata);
          end;

        ClientHeight:= min(LastItemTop + BP.Height + 10, 390);

        if (ShowModal = mrOK) then begin
          for i:= 0 to SB.ControlCount - 1 do begin
            ctl:= SB.Controls[i];
            fldno:= ctl.Tag;
            if (fldno >= 0) and (fldno < Data.Extractors.Count) then
              with TFieldExtractor(Data.Extractors[fldno]) do SetFieldBuf(pChar(pChar(msg.data) + FOffset)^, ctl, FFullType);
          end;

          if not DumpCheck.Checked or (Application.MessageBox(pChar(apublisher.dumpmessage(msg)), 'Message dump:', MB_ICONQUESTION or MB_YESNO) = ID_YES) then begin
            if assigned(Listener) then TGuiListener(Listener).doParseData(pcg_msg_data(msg)^.msg_index, msg^.data, msg^.data_size, nil, 0);
            with TGuiPublisher(apublisher) do begin
              case pcg_msg(msg)^.msg_type of
                {$ifdef with_cgate_ex}
                CG_MSG_STREAM_DATA : begin
                                       in_trs:= assigned(begin_tran_msg) and (msg_post(begin_tran_msg, 0) = CG_ERR_OK);
                                       try log(logfmtstr, [params, msg_post(pcg_msg(msg), CG_PUB_PREPARED)]);
                                       finally if in_trs and assigned(commit_tran_msg) then msg_post(commit_tran_msg, 0); end;
                                     end;
                {$endif}                     
                CG_MSG_DATA        : log(logfmtstr, [params, msg_post(pcg_msg(msg), need_reply[assigned(Listener)])]);
                else                 log('Publisher: %s; result: Unknown message type', [params]);
              end;
            end;
          end;
        end;
      finally aPublisher.msg_free(pcg_msg(msg)); end;
    finally free; end;
  end;
end;

procedure TTableForm.ListViewPopupPopup(Sender: TObject);
begin
  SendMenuItem.Enabled:= AllowSend;
  ResendMenuItem.Enabled:= AllowSend and assigned(LV.Selected);
end;

procedure TTableForm.SendMenuItemClick(Sender: TObject);
var node : tTreeNode;
begin
  node:= TreeNode.Parent;
  if assigned(node) and assigned(node.Data) then SendTransaction(TGuiPublisher(node.Data));
end;

procedure TTableForm.ResendMenuItemClick(Sender: TObject);
var node    : tTreeNode;
    msgdata : pChar;
begin
  msgdata:= nil;
  if assigned(LV.Selected) then msgdata:= Data.GetRowDataPtr(Data[LV.Selected.Index]);
  node:= TreeNode.Parent;
  if assigned(node) and assigned(node.Data) then SendTransaction(TGuiPublisher(node.Data), msgdata);
end;

end.
