{$ifndef Unix}
 {$APPTYPE CONSOLE}
{$endif} 
program repl;

uses {$ifdef FPC}
       {$ifdef MSWINDOWS}windows,{$endif}
     {$else}
       windows, 
     {$endif}
     sysutils,
     cgate;

function CHECK_FAIL(err_code: longint): boolean;
begin
  result:= (err_code = CG_ERR_OK);
  if not result then begin
    cg_log_infostr(pAnsiChar(format('Client gate error: %.8x', [err_code])));
    halt;
  end;
end;

function WARN_FAIL(err_code: longint): boolean;
begin
  result:= (err_code = CG_ERR_OK);
  if not result then cg_log_infostr(pAnsiChar(format('Client gate error: %.8x', [err_code])));
end;

function MessageCallback(hconn: THandle; hlistener: THandle; msg: pcg_msg; data: pointer): longint; cdecl;
var scheme : pcg_scheme_desc;
    msgs   : pcg_message_desc;
    flds   : pcg_field_desc;
begin
  case msg^.msg_type of
    CG_MSG_STREAM_DATA      : with pcg_msg_streamdata(msg)^ do
                                writeln(format('DATA message [table:[%d-%s], dataSize:%d]', [msg_index, msg_name, data_size]));
    CG_MSG_P2REPL_ONLINE    : cg_log_infostr('ONLINE');
    CG_MSG_TN_BEGIN         : cg_log_infostr('TN BEGIN');
    CG_MSG_TN_COMMIT        : cg_log_infostr('TN COMMIT');
    CG_MSG_OPEN             : begin
                                cg_log_infostr('OPEN');

                                scheme:= nil;
                                WARN_FAIL(cg_lsn_getscheme(hlistener, scheme));
                                if assigned(scheme) then begin
                                  msgs:= scheme^.messages;
                                  while assigned(msgs) do try
                                    cg_log_infostr(pAnsiChar(format('Message %s, block size = %d', [msgs^.name, msgs^.size])));
                                    flds:= msgs^.fields;
                                    while assigned(flds) do try
                                      cg_log_infostr(pAnsiChar(format(#$9'Field %s = %s; size:%d offset:%d', [flds^.name, flds^.field_type, flds^.size, flds^.offset])));
                                    finally flds:= flds^.next; end;
                                  finally msgs:= msgs^.next; end;
                                end;
                              end;
    CG_MSG_CLOSE            : cg_log_infostr('CLOSE');
    CG_MSG_P2REPL_LIFENUM   : cg_log_infostr(pAnsiChar(format('Life number changed to: %d', [plongint(msg^.data)^])));
    CG_MSG_P2REPL_REPLSTATE : cg_log_infostr(pAnsiChar(format('Replica state: %s', [pChar(msg^.data)])));
    else                      cg_log_infostr(pAnsiChar(format('Message 0x%.8x', [msg^.msg_type])));
  end;

  result:= CG_ERR_OK;
end;

var hConn, hListner     : THandle;
    state, process_res  : longint;
    fexit               : boolean = false;

{$ifdef MSWINDOWS}
function CtrlHandler(CtrlType: Longint): bool; stdcall;
begin
  fexit:= true;
  result:= true;
end;
{$endif}

begin
  hConn:= 0; hListner:= 0;

  {$ifdef MSWINDOWS}
  SetConsoleCtrlHandler(@CtrlHandler, true);
  {$endif}

  try

// p2tcp://127.0.0.1:4001;app_name=test1
// p2repl://FORTS_FUTINFO_REPL
// p2repl://FORTS_FUTTRADE_REPL

    CHECK_FAIL(cg_env_open('ini=repl.ini;subsystems=mq,replclient;key=11111111'));

    while not fexit do begin
      if (hConn = 0) then begin
        WARN_FAIL(cg_conn_new('p2tcp://127.0.0.1:4001;app_name=test1', hConn));
      end else begin
        WARN_FAIL(cg_conn_getstate(hConn, state));
        case state of
          CG_STATE_ERROR  : WARN_FAIL(cg_conn_close(hConn));
          CG_STATE_CLOSED : WARN_FAIL(cg_conn_open(hConn, nil));
          CG_STATE_OPENING: process_res:= cg_conn_process(hConn, 1, nil);
          CG_STATE_ACTIVE : begin
                              process_res:= cg_conn_process(hConn, 1, nil);

                              if (process_res <> CG_ERR_OK) and (process_res <> CG_ERR_TIMEOUT) then
                                cg_log_infostr(pAnsiChar(format('Warning: connection state request: %d', [process_res])));

                              if (hListner = 0) then begin
                                WARN_FAIL(cg_lsn_new(hConn,
                                                     'p2repl://FORTS_FUTINFO_REPL',
                                                     //'p2repl://FORTS_FUTTRADE_REPL;scheme=|FILE|deal.ini|FutTrade',
                                                     @MessageCallback,
                                                     nil,
                                                     hListner));
                              end else begin
                                WARN_FAIL(cg_lsn_getstate(hListner, state));
                                case state of
                                  CG_STATE_CLOSED : WARN_FAIL(cg_lsn_open(hListner, nil));
                                  CG_STATE_ERROR  : WARN_FAIL(cg_lsn_close(hListner));
                                end;
                              end;

                            end;
        end;
      end;
    end;

    if (hListner <> 0) then begin
      cg_lsn_close(hListner);
      cg_lsn_destroy(hListner);
    end;

    if (hConn <> 0) then begin
      cg_conn_close(hConn);
      cg_conn_destroy(hConn);
    end;

    CHECK_FAIL(cg_env_close);

  except on e: exception do cg_log_infostr(pAnsiChar(format('EXCEPTION: (%d) %s', [e.HelpContext, e.message]))); end;

  writeln(#$0d#$0a'PRESS ENTER TO EXIT'); readln;
end.
